@@ -22,18 +22,20 @@ end
 --- 以下は全ファイルで共有される定数
 local icflag_table = {}
 luatexja.icflag_table = icflag_table
-icflag_table.ITALIC       = 1
-icflag_table.PACKED       = 2
-icflag_table.KINSOKU      = 3
-icflag_table.FROM_JFM     = 6
--- FROM_JFM: 4, 5, 6, 7, 8 →優先度高
+icflag_table.ITALIC          = 1
+icflag_table.PACKED          = 2
+icflag_table.KINSOKU         = 3
+icflag_table.FROM_JFM        = 6
+-- FROM_JFM: 4, 5, 6, 7, 8 →優先度高（伸びやすく，縮みやすい）
 -- 6 が標準
-icflag_table.KANJI_SKIP   = 9
-icflag_table.XKANJI_SKIP  = 10
-icflag_table.PROCESSED    = 11
-icflag_table.IC_PROCESSED = 12
-icflag_table.BOXBDD       = 15
-icflag_table.PROCESSED_BEGIN_FLAG = 32
+icflag_table.KANJI_SKIP      = 9
+icflag_table.KANJI_SKIP_JFM  = 10
+icflag_table.XKANJI_SKIP     = 11
+icflag_table.XKANJI_SKIP_JFM = 12
+icflag_table.PROCESSED       = 13
+icflag_table.IC_PROCESSED    = 14
+icflag_table.BOXBDD          = 15
+icflag_table.PROCESSED_BEGIN_FLAG = 128
 
 local stack_table_index = {}
 luatexja.stack_table_index = stack_table_index
@@ -41,7 +43,9 @@ stack_table_index.PRE  = 0x200000 -- characterごと
 stack_table_index.POST = 0x400000 -- characterごと
 stack_table_index.KCAT = 0x600000 -- characterごと
 stack_table_index.XSP  = 0x800000 -- characterごと
-stack_table_index.JWP  = 0 -- 0のみ
+stack_table_index.JWP  = 0 -- これだけ
+stack_table_index.KSK  = 1 -- これだけ
+stack_table_index.XSK  = 2 -- これだけ
 stack_table_index.MJT  = 0x100 -- 0--255
 stack_table_index.MJS  = 0x200 -- 0--255
 stack_table_index.MJSS = 0x300 -- 0--255
@@ -51,9 +55,46 @@ local userid_table = {}
 luatexja.userid_table = userid_table
 userid_table.IHB  = luatexbase.newuserwhatsitid('inhibitglue',  'luatexja') -- \inhibitglue
 userid_table.STCK = luatexbase.newuserwhatsitid('stack_marker', 'luatexja') -- スタック管理
-userid_table.OTF  = luatexbase.newuserwhatsitid('char_by_cid',  'luatexja') -- luatexja-otf
 userid_table.BPAR = luatexbase.newuserwhatsitid('begin_par',    'luatexja') -- 「段落始め」
 
+------------------------------------------------------------------------
+-- FIX node.remove
+------------------------------------------------------------------------
+do
+   local node_remove, node_next, node_prev = node.remove, node.next, node.prev
+   function luatexja.node_remove (head, current)
+      if head==current then
+         local q, r = node_next(current), node_prev(current)
+         if q then q.prev = r end
+         if r and node_next(r)==current then 
+            r.next = q
+         end
+         return q, q
+      else
+         return node_remove(head, current)
+      end
+   end
+   local Dnode = node.direct or node
+   if Dnode~=node then
+      local Dnode_remove, setfield = Dnode.remove, Dnode.setfield
+      local Dnode_next, Dnode_prev = Dnode.getnext, Dnode.getprev
+      function luatexja.Dnode_remove (head, current)
+         if head==current then
+            local q, r = Dnode_next(current), Dnode_prev(current)
+            if q then setfield(q, 'prev', r) end
+            if r and Dnode_next(r) == current then
+               setfield(r, 'next', q)
+            end
+            return q, q
+         else
+            return Dnode_remove(head, current)
+         end
+      end
+   else
+      luatexja.Dnode_remove = luatexja.node_remove 
+   end
+end
+
 --- 定義終わり
 
 local load_module = luatexja.load_module
@@ -76,10 +117,6 @@ local attr_icflag = luatexbase.attributes['ltj@icflag']
 local attr_uniqid = luatexbase.attributes['ltj@uniqid']
 local cat_lp = luatexbase.catcodetables['latex-package']
 
-
----- table: charprop_stack_table [stack_level].{pre|post|xsp}[chr_code]
-
-
 -- Three aux. functions, bollowed from tex.web
 
 local unity=65536
@@ -132,83 +169,95 @@ end
 ------------------------------------------------------------------------
 
 -- EXT: print parameters that don't need arguments
-function luatexja.ext_get_parameter_unary(k)
-   local t = tex.getcount('ltj@@stack')
-   if k == 'yalbaselineshift' then
-      tex.write(print_scaled(tex.getattribute('ltj@yablshift'))..'pt')
-   elseif k == 'yjabaselineshift' then
-      tex.write(print_scaled(tex.getattribute('ltj@ykblshift'))..'pt')
-   elseif k == 'kanjiskip' then
-      tex.write(print_spec(ltjs.get_skip_table('kanjiskip', t)))
-   elseif k == 'xkanjiskip' then
-      tex.write(print_spec(ltjs.get_skip_table('xkanjiskip', t)))
-   elseif k == 'jcharwidowpenalty' then
-      tex.write(ltjs.get_penalty_table(stack_table_index.JWP, 0, t))
-   elseif k == 'autospacing' then
-      tex.write(tex.getattribute('ltj@autospc'))
-   elseif k == 'autoxspacing' then
-      tex.write(tex.getattribute('ltj@autoxspc'))
-   elseif k == 'differentjfm' then
-      if luatexja.jfmglue.diffmet_rule == math.max then
-	 tex.write('large')
-      elseif luatexja.jfmglue.diffmet_rule == math.min then
-	 tex.write('small')
-      elseif luatexja.jfmglue.diffmet_rule == math.two_average then
-	 tex.write('average')
-      elseif luatexja.jfmglue.diffmet_rule == math.two_paverage then
-	 tex.write('paverage')
-      elseif luatexja.jfmglue.diffmet_rule == math.two_pleft then
-	 tex.write('pleft')
-      elseif luatexja.jfmglue.diffmet_rule == math.two_pright then
-	 tex.write('pright')
-      elseif luatexja.jfmglue.diffmet_rule == math.two_add then
-	 tex.write('both')
-      else -- This can't happen.
-	 tex.write('???')
+do
+   luatexja.unary_pars = {
+      yalbaselineshift = function(t) 
+	 return print_scaled(tex.getattribute('ltj@yablshift'))..'pt'
+      end,
+      yjabaselineshift = function(t) 
+	 return print_scaled(tex.getattribute('ltj@ykblshift'))..'pt'
+      end,
+      kanjiskip = function(t) 
+	 return print_spec(ltjs.get_stack_skip(stack_table_index.KSK, t))
+      end,
+      xkanjiskip = function(t) 
+	 return print_spec(ltjs.get_stack_skip(stack_table_index.XSK, t))
+      end,
+      jcharwidowpenalty = function(t)
+	 return ltjs.get_stack_table(stack_table_index.JWP, 0, t)
+      end,
+      autospacing = function(t)
+	 return tex.getattribute('ltj@autospc')
+      end,
+      autoxspacing = function(t)
+	 return tex.getattribute('ltj@autoxspc')
+      end,
+      differentjfm = function(t)
+	 local f, r = luatexja.jfmglue.diffmet_rule, '???'
+	 if f == math.max then r = 'large'
+	 elseif f == math.min then r = 'small'
+	 elseif f == math.two_average then r = 'average'
+	 elseif f == math.two_paverage then r = 'paverage'
+	 elseif f == math.two_pleft then r = 'pleft'
+	 elseif f == math.two_pright then r = 'pright'
+	 elseif f == math.two_add then r = 'both'
+	 end
+	 return r
+      end
+   }
+
+   local unary_pars = luatexja.unary_pars
+   function luatexja.ext_get_parameter_unary(k)
+      if unary_pars[k] then
+	 tex.write(tostring(unary_pars[k](tex.getcount('ltj@@stack'))))
       end
    end
 end
 
 
 -- EXT: print parameters that need arguments
-function luatexja.ext_get_parameter_binary(k,c)
-   local t = tex.getcount('ltj@@stack')
-   if type(c)~='number' then
-      ltjb.package_error('luatexja',
-			 'invalid the second argument (' .. tostring(c) .. ')',
-			 'I changed this one to zero.')
-      c=0
-   end
-   if k == 'jacharrange' then
-      if c>=31*ltjc.ATTR_RANGE then 
-	 ltjb.package_error('luatexja',
-			    'invalid character range number (' .. c .. ')',
-			    'A character range number should be in the range 0..'
-                               .. 31*ltjc.ATTR_RANGE-1 .. ",\n"..
-			     'So I changed this one to zero.')
-	 c=0
-      end
+do
+   luatexja.binary_pars = {
+      jacharrange = function(c, t)
+	 if type(c)~='number' or c<0 or c>31*ltjc.ATTR_RANGE then
+	    -- 0 はエラーにしない（隠し）
+	    ltjb.package_error('luatexja',
+			       'invalid character range number (' .. tostring(c) .. ')',
+			       'A character range number should be in the range 1..'
+				  .. 31*ltjc.ATTR_RANGE .. ",\n"..
+				  'So I changed this one to ' .. 31*ltjc.ATTR_RANGE .. ".")
+	    c=0 -- external range 217 == internal range 0
+	 elseif c==31*ltjc.ATTR_RANGE then c=0
+	 end
       -- 負の値は <U+0080 の文字の文字範囲，として出てくる．この時はいつも欧文文字なので 1 を返す
-      tex.write( (c<0) and -1 or ltjc.get_range_setting(c))
-   else
-      if c<0 or c>0x10FFFF then
-	 ltjb.package_error('luatexja',
-			    'bad character code (' .. c .. ')',
-			    'A character number must be between -1 and 0x10ffff.\n'..
-			       "(-1 is used for denoting `math boundary')\n"..
-			       'So I changed this one to zero.')
-	 c=0
-      end
-      if k == 'prebreakpenalty' then
-	 tex.write(ltjs.get_penalty_table(stack_table_index.PRE + c, 0, t))
-      elseif k == 'postbreakpenalty' then
-	 tex.write(ltjs.get_penalty_table(stack_table_index.POST+ c, 0, t))
-      elseif k == 'kcatcode' then
-	 tex.write(ltjs.get_penalty_table(stack_table_index.KCAT+ c, 0, t))
-      elseif k == 'chartorange' then 
-	 tex.write(ltjc.char_to_range(c))
-      elseif k == 'jaxspmode' or k == 'alxspmode' then
-	 tex.write(ltjs.get_penalty_table(stack_table_index.XSP + c, 3, t))
+	 return (c<0) and 1 or ltjc.get_range_setting(c)
+      end,
+      prebreakpenalty = function(c, t)
+	 return ltjs.get_stack_table(stack_table_index.PRE 
+					  + ltjb.in_unicode(c, true), 0, t)
+      end,
+      postbreakpenalty = function(c, t)
+	 return ltjs.get_stack_table(stack_table_index.POST 
+					  + ltjb.in_unicode(c, true), 0, t)
+      end,
+      kcatcode = function(c, t)
+	 return ltjs.get_stack_table(stack_table_index.KCAT 
+					  + ltjb.in_unicode(c, false), 0, t)
+      end,
+      chartorange = function(c, t)
+	 return ltjc.char_to_range(ltjb.in_unicode(c, false))
+      end,
+      jaxspmode = function(c, t)
+	 return ltjs.get_stack_table(stack_table_index.XSP
+					  + ltjb.in_unicode(c, true), 3, t)
+      end,
+   }
+   local binary_pars = luatexja.binary_pars 
+
+   binary_pars.alxspmode = binary_pars.jaxspmode
+   function luatexja.ext_get_parameter_binary(k,c)
+      if binary_pars[k] then
+	 tex.write(tostring(binary_pars[k](c,tex.getcount('ltj@@stack'))))
       end
    end
 end
@@ -219,41 +268,49 @@ function luatexja.ext_print_global()
 end
 
 -- main process
--- mode = true iff main_process is called from pre_linebreak_filter
-local function main_process(head, mode, dir)
-   local p = head
-   p = ltjj.main(p,mode)
-   if p then p = ltjw.set_ja_width(p, dir) end
-   return p
+do
+   local Dnode = node.direct or node
+   local nullfunc = function (n) return n end
+   local to_node = (Dnode ~= node) and Dnode.tonode or nullfunc
+   local to_direct = (Dnode ~= node) and Dnode.todirect or nullfunc
+   -- mode = true iff main_process is called from pre_linebreak_filter
+   local function main_process(head, mode, dir)
+      local p = to_direct(head)
+      p = ltjj.main(p,mode)
+      if p then p = ltjw.set_ja_width(p, dir) end
+      return to_node(p)
+   end
+   
+   -- callbacks
+   
+   luatexbase.add_to_callback(
+      'pre_linebreak_filter', 
+      function (head,groupcode)
+	 return main_process(head, true, tex.textdir)
+      end,'ltj.pre_linebreak_filter',
+      luatexbase.priority_in_callback('pre_linebreak_filter',
+				      'luaotfload.node_processor') + 1)
+   luatexbase.add_to_callback(
+      'hpack_filter', 
+      function (head,groupcode,size,packtype, dir)
+	 return main_process(head, false, dir)
+      end,'ltj.hpack_filter',
+      luatexbase.priority_in_callback('hpack_filter',
+				      'luaotfload.node_processor') + 1)
 end
 
--- callbacks
-
-luatexbase.add_to_callback('pre_linebreak_filter', 
-   function (head,groupcode)
-      return main_process(head, true, tex.textdir)
-   end,'ltj.pre_linebreak_filter',
-   luatexbase.priority_in_callback('pre_linebreak_filter',
-				   'luaotfload.node_processor') + 1)
-luatexbase.add_to_callback('hpack_filter', 
-  function (head,groupcode,size,packtype, dir)
-     return main_process(head, false, dir)
-  end,'ltj.hpack_filter',
-   luatexbase.priority_in_callback('hpack_filter',
-				   'luaotfload.node_processor') + 1)
-
 -- define_font
-
-local otfl_fdr = fonts.definers.read
-function luatexja.font_callback(name, size, id)
-  return ltjf.font_callback(
-     name, size, id, 
-     function (name, size, id) return ltjr.font_callback(name, size, id, otfl_fdr) end
-  )
+do
+   local otfl_fdr = fonts.definers.read
+   local ltjr_font_callback = ltjr.font_callback
+   function luatexja.font_callback(name, size, id)
+      return ltjf.font_callback(
+	 name, size, id, 
+	 function (name, size, id) return ltjr_font_callback(name, size, id, otfl_fdr) end
+      )
+   end
+   luatexbase.add_to_callback('define_font',luatexja.font_callback,"luatexja.font_callback", 1)
 end
---luatexbase.remove_from_callback('define_font',"luaotfload.define_font")
-luatexbase.add_to_callback('define_font',luatexja.font_callback,"luatexja.font_callback", 1)
-
 
 
 
@@ -315,11 +372,9 @@ local function debug_show_node_X(p,print_fn)
       if get_attr_icflag(p) == icflag_table.PACKED then
          s = s .. ' (packed)'
       end
-      print_fn(s)
-      local q = p.head
-      debug_depth=debug_depth.. '.'
-      while q do 
-         debug_show_node_X(q, print_fn); q = node_next(q)
+      print_fn(s); debug_depth=debug_depth.. '.'
+      for q in node.traverse(p.head) do
+         debug_show_node_X(q, print_fn)
       end
       debug_depth=k
    elseif pt == 'glue' then
@@ -329,8 +384,12 @@ local function debug_show_node_X(p,print_fn)
          s = s .. ' (from JFM: priority ' .. get_attr_icflag(p)-icflag_table.FROM_JFM .. ')'
       elseif get_attr_icflag(p)==icflag_table.KANJI_SKIP then
 	 s = s .. ' (kanjiskip)'
+      elseif get_attr_icflag(p)==icflag_table.KANJI_SKIP_JFM then
+	 s = s .. ' (kanjiskip, JFM specified)'
       elseif get_attr_icflag(p)==icflag_table.XKANJI_SKIP then
 	 s = s .. ' (xkanjiskip)'
+      elseif get_attr_icflag(p)==icflag_table.XKANJI_SKIP_JFM then
+	 s = s .. ' (xkanjiskip, JFM specified)'
       end
       print_fn(s)
    elseif pt == 'kern' then
@@ -387,10 +446,13 @@ local function debug_show_node_X(p,print_fn)
    elseif pt=='math_char' then
       s = base .. ' fam: ' .. p.fam .. ' , char = ' .. utf.char(p.char)
       print_fn(s)
-   elseif pt=='sub_box' then
+   elseif pt=='sub_box' or pt=='sub_mlist' then
       print_fn(base)
       if p.head then
-         debug_depth = k .. '.'; debug_show_node_X(p.head, print_fn); 
+         debug_depth = k .. '.'; 
+	 for q in node.traverse(p.head) do
+	    debug_show_node_X(q, print_fn)
+	 end
       end
    else
       print_fn(base)
